/*  Written by Julian Grizzard
 *  September 2, 2002
 *  Contact: gtg993j@mail.gatech.edu
 *
 *  Please contact me if you decide to publish any of this code.  Thanks!
 */
#include <stdio.h>
#include <stdlib.h>
#include "ai.h"

/* calc the shortest path from startX, startY to endX, endY */
list * getPathInit(s_world *world, int startX, int startY, int endX, int endY, 
                   list *path, int *count);

/* helper to calc the shortest path from startX, startY to endX, endY */
list * getPath(s_world *world, int startX, int startY, int endX, int endY, 
            list *path, int *count);

/* returns int greater than 0 if you should not go to those
 * coordinates
 * returns 0 if coordinates clear
 */
int dontGo(char **map, int x, int y, list *path);

/* frees the memory and all elements in path list */
void freePath(list *path);

/* prints the path */
void printPath(list *path);

/* creates a character list of items to be dropped/picked */
char * makeList(list *thePkgs);

/* takes next move in course if possible */
int moveNext(s_world *world, list *robotL, list *homeL, list *pkgL, 
             s_robot *meBot);

/* returns the bid that should be used for this action */
int computePriority(s_world *world, list *robotL, list *homeL, list *pkgL,
                    char actionType);

/* returns NULL or list of pkgs to be picked up */
list * findPkgs(s_robot *meBot, list *pkgL);

/* finds biggest pkg */
s_pkg * findBigPkg(list *pkgL);

/* returns total weight of packages in list */
int weightPkgList(list *thePkgs);

int initialize();

int init = 0;
s_square **squares = NULL;
int **dstM = NULL;
list *course;
element *homeItr;

int issueCommand(s_world *world, list *robotL, list *homeL, list *pkgL)
{
   element *pkItr, *botItr, *pathItr;
   char *arguments;
   s_robot *meBot;
   //s_square *theSquare;
   s_pkg *thePkg;
   //list *thePath;
   list *pkUpPkgL;
   int count;
   headElement(robotL, &botItr);
   meBot = botItr->value;
   pathItr = NULL;

   if(init == 0)
   {
      /* first time running issueCommand, initialize everything */
      printf("Initializing AI...\n");
      initialize(world);
   }

/*
   printf("getting path...\n"); 
   thePath = newList();
   count = 100000;
   thePath = getPathInit(world, 15, 1, 1, 1, thePath, &count);
   if(thePath == NULL)
   {
      printf("Path not found!\n");
   }
   else
   {
      printPath(thePath);
   }
   exit(0);
*/


   /* First off, if we are holding any packages that need to 
    * be dropped off, then by all means drop 'em with a boOOT
    */
   if(headElement(meBot->pkgs, &pkItr) >= 0)
   {
      list *toBeDropped = newList();
      thePkg = NULL;
      do
      {
         thePkg = pkItr->value;
         if(thePkg->x == meBot->x && thePkg->y == meBot->y)
         {
            /* found one, add it to the list of to be dropped off */
            addElement(toBeDropped, thePkg);
         }
      } while(nextElement(meBot->pkgs, pkItr, &pkItr) >= 0);
      if(listSize(toBeDropped) > 0)
      {
         /* Drop 'em */
         arguments = makeList(toBeDropped);
         drop(arguments, 
              computePriority(world, robotL, homeL, pkgL, A_DROP),
              robotL);
         free(arguments);
         return 0;
      }
   } /* else we don't have any packages on us */

   /* Second, if we are standing on any packages, then pick 
    * them up.
    */
   pkUpPkgL = findPkgs(meBot, pkgL);
   if(pkUpPkgL != NULL)
   {
      /* Pick them up */
      arguments = makeList(pkUpPkgL);
      pick(arguments,
           computePriority(world, robotL, homeL, pkgL, A_PICK),
           robotL);
      free(arguments);
      return 0;   
   }

   /* Third, if we are on a course, then move to next position
    */
   if(moveNext(world, robotL, homeL, pkgL, meBot) == 0)
      return 0;

   /* Fourth, if we are holding any packages, then set a course
    * for the biggest one
    */
   thePkg = findBigPkg(meBot->pkgs);
   if(thePkg != NULL)
   {
      count = 100000;
      freePath(course);
      course = getPathInit(world, meBot->x, meBot->y, thePkg->x, 
                           thePkg->y, newList(), &count);
      if(moveNext(world, robotL, homeL, pkgL, meBot) == 0)
         return 0;
   }
 

   /* Fifth, If we know of any packages, set a course for the biggest
    * one
    */
   thePkg = findBigPkg(pkgL);
   if(thePkg != NULL)
   {
      count = 100000;
      freePath(course);
      course = getPathInit(world, meBot->x, meBot->y, thePkg->x, 
                           thePkg->y, newList(), &count);
      if(moveNext(world, robotL, homeL, pkgL, meBot) == 0)
         return 0;
   }

   /* Sixth, Set a course for next home to visit
    */
   if(homeItr == NULL)
      headElement(homeL, &homeItr);
   if(homeItr != NULL)
   {
      s_home *theHome;
      theHome = (s_home *) homeItr->value;
      count = 100000;
      freePath(course);
      //printf("HOME X %d Y %d\n", theHome->x, theHome->y);
      course = getPathInit(world, meBot->x, meBot->y, theHome->x, 
                           theHome->y, newList(), &count);
      nextElement(homeL, homeItr, &homeItr);
      if(moveNext(world, robotL, homeL, pkgL, meBot) == 0)
         return 0;
   }

   /* Default in case everything goes wrong, we won't be a 
    * complete embarressment (assuming we get here and don't crash)
    */
   pick(" 10", 1, robotL);
   return 0;
}

s_pkg * findBigPkg(list *pkgL)
{
   element *pkgItr;
   element *pkgFing;
   if(pkgL == NULL)
   {
      return NULL;
   }
   headElement(pkgL, &pkgItr);
   pkgFing = pkgItr;
   while(pkgItr != NULL)
   {
      nextElement(pkgL, pkgItr, &pkgItr);
      if(pkgItr != NULL)
      {
         if(((s_pkg *)pkgItr->value)->weight >
            ((s_pkg *)pkgFing->value)->weight)
         {
            pkgFing = pkgItr;
         }
      }
   }
   if(pkgFing == NULL)
   {
      return NULL;
   }
   return (s_pkg *)pkgFing->value;
}

list * findPkgs(s_robot *meBot, list *pkgL)
{
   list *thePkgs = NULL;
   element *pkgLItr;
   if(pkgL == NULL) return NULL;
   headElement(pkgL, &pkgLItr);
   while(pkgLItr != NULL)
   {
      s_pkg *thePkg = (s_pkg *) pkgLItr->value;
      /* if it matches our location and
       * it is not already carried and
       * we have room for it
       */
      if(thePkg->xLoc == meBot->x && thePkg->yLoc == meBot->y &&
         thePkg->state == P_GROUNDED &&
         thePkg->weight <= meBot->cpty - weightPkgList(thePkgs))
      {
         if(thePkgs == NULL) thePkgs = newList();
         addElement(thePkgs, thePkg);
      }
      nextElement(pkgL, pkgLItr, &pkgLItr);
   }
   return thePkgs;
}

int weightPkgList(list *thePkgs)
{
   element *thePkg = NULL;
   int weight = 0;
   if(thePkgs == NULL)
   {
      return 0;
   }
   headElement(thePkgs, &thePkg);
   if(thePkg == NULL)
   {
      return 0;
   }
   while(thePkg != NULL)
   {
      weight = weight + ((s_pkg *)thePkg->value)->weight;
      nextElement(thePkgs, thePkg, &thePkg);
   }
   return weight;
}

int moveNext(s_world *world, list *robotL, list *homeL, list *pkgL, 
             s_robot *meBot)
{
   element *pathItr;
   s_square *theSquare;
   if(headElement(course, &pathItr) < 0)
      return -1;
   if(pathItr == NULL)
      return -1;
   theSquare = (s_square *) pathItr->value;
   if(theSquare->x == meBot->x && theSquare->y == meBot->y && 
      listSize(course) > 1)
   {
      /* We are on a path */
      removeElement(course, theSquare);
      headElement(course, &pathItr);
      if(pathItr == NULL)
         return -1;
      theSquare = (s_square *) pathItr->value;
      if(theSquare->x > meBot->x)
      {
         move(A_EAST_S,
              computePriority(world, robotL, homeL, pkgL, A_MOVE),
              robotL);
         return 0;
      }
      else if(theSquare->x < meBot->x)
      {
         move(A_WEST_S,
              computePriority(world, robotL, homeL, pkgL, A_MOVE),
              robotL);
         return 0;
      }
      else if(theSquare->y > meBot->y)
      {
         move(A_NORTH_S,
              computePriority(world, robotL, homeL, pkgL, A_MOVE),
              robotL);
         return 0;
      }
      else if(theSquare->y < meBot->y)
      {
         move(A_SOUTH_S,
              computePriority(world, robotL, homeL, pkgL, A_MOVE),
              robotL);
         return 0;
      }
     
   }
   return -1;
}

int initialize(s_world *world)
{
   int i;
   course = newList();
   squares = (s_square **) malloc((world->width + 2) * sizeof(s_square *));
   dstM = (int **) malloc((world->width + 2) * sizeof(int *));
   for(i = 0; i < world->width + 2; i++)
   {
      squares[i] = (s_square *) malloc((world->height + 2) * sizeof(s_square));
      dstM[i] = (int *) malloc((world->height + 2) * sizeof(int));
   }
   for(i = world->height + 1; i >= 0; i--)
   {
      int j = 0;
      for(j = 0; j < world->width + 2; j++)
      {
         squares[j][i] = *newSquare(j, i);
         dstM[j][i] = 1000000;
      }
   }
   init = 1;
   return 0;
}

char * makeList(list *thePkgs)
{
   char *theList;
   char *finger;
   element *pkgItr;
   
   theList = (char *) malloc(20 * listSize(thePkgs));
   finger = theList;

   headElement(thePkgs, &pkgItr);

   do
   {
      int count = 0;
      s_pkg *thePkg = pkgItr->value;
      count = sprintf(finger, " %d", thePkg->id); 
      finger = finger + count;
   } while(nextElement(thePkgs, pkgItr, &pkgItr) >= 0);

   return theList;
}

/* returns the bid that should be used for this action */
int computePriority(s_world *world, list *robotL, list *homeL, list *pkgL,
                    char actionType)
{
   //s_robot *meBot = NULL;
   //headElement(robot, 
   return 1;
}

list * getPathInit(s_world *world, int startX, int startY, int endX, int endY, 
                   list *path, int *count)
{
   list *thePath;
   element *finger1, *finger2, *finger3;
   int i, j = 0;
   for(i = 0; i < world->width + 2; i++)
   {
      for(j = 0; j < world->height + 2; j++)
      {
         dstM[i][j] = 1000000;
      }
   }
   thePath = getPath(world, startX, startY, endX, endY, path, count);
   if(thePath == NULL)
   {
      return NULL;
   }
   headElement(thePath, &finger1);
   while(finger1 != NULL)
   {
      s_square *s1, *s2;
      s1 = finger1->value;
      nextElement(thePath, finger1, &finger2);
      while(finger2 != NULL)
      {
         s2 = finger2->value;
         if((s1->x == s2->x && s1->y + 1 == s2->y) ||
            (s1->x == s2->x && s1->y - 1 == s2->y) ||
            (s1->x + 1 == s2->x && s1->y == s2->y) ||
            (s1->x - 1 == s2->x && s1->y == s2->y))
         {
            /* We found a short cut, unless we are just
             * looking at the next one in the list 
             *
             * so delete the extra road!
             */
            finger3 = finger1;
            while(finger3->next != finger2)
            {
               removeNext(thePath, finger3); 
            }
         }
         nextElement(thePath, finger2, &finger2);
      }
      nextElement(thePath, finger1, &finger1);
   }
   return thePath;
}

list * getPath(s_world *world, int startX, int startY, int endX, int endY, 
            list *path, int *count)
{
   char **map = world->map;
   list *localPath, *pathNorth, *pathSouth, *pathEast, *pathWest = NULL;
   int countNorth, countSouth, countEast, countWest = 0;
   int size = listSize(path);
   if(size >= *count)
   {
      /* this path will not be any shorter so go ahead and return
       * NULL
       */
      //printf("catch guard for %d %d says too long\n", startX, startY);
      return NULL;
   }
   if(size >= dstM[startX][startY])
   {
      /* Already found a path to this point that was same or less
       * distance from start point so return NULL
       */
      //printf("catch guard for %d %d says ALREADY BEEN HERE\n", startX, startY);
      return NULL;
   }
   else
   {
      dstM[startX][startY] = 0;
   }
   localPath = copyList(path);
   addElement(localPath, &squares[startX][startY]);
   //printf("sx: %d sy: %d ex: %d ey: %d\n", startX, startY, endX, endY);
   //printPath(localPath); 
   /* FOUND a path wooHOO! */
   if(startX == endX && startY == endY)
   {
      //printf("FOUND PATH WOHOO! - ");
      if(size < *count)
      {
         *count = size;
      }
      //printPath(localPath);
      return localPath;
   }
   /* If we cannot go in a direction because of a wall, water,
    * or because we've already been in all those directions
    * then, then don't recurse.
    */
   if(dontGo(map, startX, startY+1, localPath))
   {
      pathNorth = NULL;
      countNorth = -1;
   }
   else
   {
      //printf("go north\n");
      pathNorth = getPath(world, startX, startY+1, endX, endY, localPath, 
                          count);
      if(pathNorth != NULL) 
         countNorth = listSize(pathNorth);
      else
         countNorth = -1;
   }
   if(pathNorth != NULL) return pathNorth;

   if(dontGo(map, startX, startY-1, localPath))
   {
      pathSouth = NULL;
      countSouth = -1;
   }
   else
   {
      //printf("go south\n");
      pathSouth = getPath(world, startX, startY-1, endX, endY, localPath,
                          count);
      if(pathSouth != NULL) 
         countSouth = listSize(pathSouth);
      else
         countSouth = -1;
   }
   if(pathSouth != NULL) return pathSouth;

   if(dontGo(map, startX+1, startY, localPath))
   {
      pathEast = NULL;
      countEast = -1;
   }
   else
   {
      //printf("go east\n");
      pathEast = getPath(world, startX+1, startY, endX, endY, localPath,
                         count);
      if(pathEast != NULL) 
         countEast = listSize(pathEast);
      else
         countEast = -1;
   }
   if(pathEast != NULL) return pathEast;

   if(dontGo(map, startX-1, startY, localPath))
   {
      pathWest = NULL;
      countWest = -1;
   }
   else
   {
      //printf("go west\n");
      pathWest = getPath(world, startX-1, startY, endX, endY, localPath,
                         count);
      if(pathWest != NULL) 
         countWest = listSize(pathWest);
      else
         countWest = -1;
   }
   if(pathWest != NULL) return pathWest;

   //printf("NORTH: "); 
   //printPath(pathNorth);
   //printf("SOUTH: "); 
   //printPath(pathSouth);
   //printf("EAST: "); 
   //printPath(pathEast);
   //printf("WEST: ");
   //printPath(pathWest);
   if(countNorth == -1 &&  countSouth == -1 && countEast == -1 &&
      countWest == -1)
   {
      freePath(localPath);
      //printf("Reached a DEAD END!\n");
      return NULL;
   }
   else if(countNorth >= 1 && 
           (countSouth <= 0 || countNorth <= countSouth) && 
           (countEast <= 0 || countNorth <= countEast) && 
           (countWest <= 0 || countNorth <= countWest))
   {
      //printf("North is best pick from here(%d, %d)! - ", startX, startY);
      freePath(pathSouth);
      freePath(pathEast);
      freePath(pathWest);
      //printPath(pathNorth);
      return pathNorth;
   }
   else if(countSouth >= 1 &&  
           (countEast <= 0 || countSouth <= countEast) && 
           (countWest <= 0 || countSouth <= countWest))
   {
      //printf("South is best pick from here(%d, %d)! - ", startX, startY);
      freePath(pathNorth);
      freePath(pathEast);
      freePath(pathWest);
      //printPath(pathSouth);
      return pathSouth;
   }
   else if(countEast >= 1 && 
           (countWest <= 0 || countEast <= countWest))
   {
      //printf("East is best pick from here(%d, %d)! - ", startX, startY);
      freePath(pathNorth);
      freePath(pathSouth);
      freePath(pathWest);
      //printPath(pathEast);
      return pathEast;
   }
   else
   {
      //printf("West is best pick from here(%d, %d)! - ", startX, startY);
      freePath(pathNorth);
      freePath(pathSouth);
      freePath(pathEast);
      ////printPath(pathWest);
      return pathWest;
   }

}

int dontGo(char **map, int x, int y, list *path)
{
   element *pathItr;
   ////printf("don't go computer for xy: %d %d\n", x, y);
   if(map[x][y] == B_WALL || map[x][y] == B_WATER)
   {
      //printf("says there is a wall\n");
      return 1;
   }
   if(path != NULL)
   {
      headElement(path, &pathItr);
      while(pathItr != NULL)
      {
         s_square *theSquare = pathItr->value;
         if(theSquare->x == x && theSquare->y == y)
         {
            //printf("says already been there\n");
            return 1;
         }
         nextElement(path, pathItr, &pathItr);
      }
   }
   //printf("says good to go\n");
   return 0;
}

void freePath(list *path)
{
   freeList(path);
/*
   element *finger;
   if(path == NULL)
   {
      return;
   }
   headElement(path, &finger);
   while(finger != NULL)
   {
      free(finger->value);
      nextElement(path, finger, &finger);
   }
   freeList(path);
*/
}

void printPath(list *thePath)
{
   element *pathItr;
   printf("printing path...");
   if(thePath == NULL)
   {
      printf("NO PATH\n");
      return;
   }
   headElement(thePath, &pathItr);
   if(pathItr == NULL)
   {
      printf("NO PATH\n");
      return;
   }
   while(pathItr != NULL)
   {
      s_square *theSquare = pathItr->value;
      printf(" (%d, %d)", theSquare->x, theSquare->y);
      nextElement(thePath, pathItr, &pathItr);
   }
   printf("\n");
}
