/*  Written by Julian Grizzard
 *  September 2, 2002
 *  Contact: gtg993j@mail.gatech.edu
 *
 *  Please contact me if you decide to publish any of this code.  Thanks!
 */

#ifndef _CLIENT_H
#define _CLIENT_H

#include "robot.h"
#include "world.h"
#include "linkedList.h"


/* Constants */

#define MAX_HOST_NAME 255
#define MAX_RECV_MSG 2000000

/* Errors */
#define EBAD_HOSTNAME 200
#define ESOCKET 201
#define EBIND 202
#define ECONNECT 203
#define ESEND 204
#define ERECV 205
#define EBAD_MAP 206
#define EBAD_CONF 207
#define EBAD_WORLD 208
#define EBAD_PACK 209

/* Available functions */

/* Initializes the world, does initialize phase
 * 1) Client -> "Player"
 * 2) Server <- Board
 * 3) Server <- Player config
 * 4) Server <- Init Response
 */
int initializeWorld(s_world *world, list *robotL, list *homeL, 
                    list *pkgL, char *hostName, int port);

/* Gets List of packages
 * - Step 1 of Game play ID X Y WEIGHT ID X Y WEIGHT...
 */
int getPackages(list *pkgL, list *robotL);

/* Sends command
 * - Step 2 of Game play BID Move|Pick|Drop [ID ID...]
 */
int sendCommand(char *command);

/* Gets server response during game play
 * - Step 3 of Game play #ID ACTIONS #ID ACTIONS...
 */
int updateWorld(s_world *world, list *robotL, list *homeL, list *pkgL);

#endif
