/*  Written by Julian Grizzard
 *  September 2, 2002
 *  Contact: gtg993j@mail.gatech.edu
 *
 *  Please contact me if you decide to publish any of this code.  Thanks!
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "linkedList.h"

list *newList()
{
   list *newL = (list *) malloc(sizeof(list));
   newL->head = NULL;
   newL->size = 0;
   return newL;
}

list *copyList(list *theList)
{
   list *newL;
   element *finger;
   if(theList == NULL)
   {
      return NULL;
   }
   newL = newList();
   headElement(theList, &finger);
   while(finger != NULL)
   {
      addElement(newL, finger->value);
      nextElement(newL, finger, &finger);
   }
   return newL;
}

int freeList(list *theList)
{
   element *finger;
   element *next;
   if(theList == NULL)
   {
      errno = EBAD_LIST;
      return -1;
   }
   headElement(theList, &finger);
   while(finger != NULL)
   {
      next = finger->next;
      free(finger);
      finger = next;
   }
   free(theList);
   return 0;
}

int addElement(list *theList, void *value)
{
   element *finger;
   element *newElem;
   if(theList == NULL)
   {
      errno = EBAD_LIST;
      return -1;
   }

   newElem = (element *) malloc(sizeof(newElem));
   newElem->value = value;
   newElem->next = NULL;

   if(theList->head == NULL)
   {
      theList->head = newElem;
      theList->size++;
      return 0;
   }

   /* there is at least one element in the list */
   finger = theList->head;
   while(finger->next != NULL)
   {
      if(finger->value == value)
      {
         free(newElem);
         return 0;
      }
      finger = finger->next;
   } /* finger points to tail of list */
  
   finger->next = newElem;
   theList->size++;

   return 0;
}

int removeElement(list *theList, void *value)
{
   element *finger;
   element *temp;
   if(theList == NULL)
   {
      errno = EBAD_LIST;
      return -1;
   }
   
   if(theList->head == NULL)
   {
      /* not found */
      return -1;
   }
   else if(theList->head->value == value)
   {
      /* found at head */
      finger = theList->head;
      theList->head = theList->head->next;
      theList->size--;
      free(finger);
      return 0;
   }

   finger = theList->head;
   while(finger->next != NULL)
   {
      if(finger->next->value == value)
      {
         /* found at next elem */
         temp = finger->next;
         finger->next = finger->next->next;
         theList->size--;
         free(temp);
         return 0;
      } 
      finger = finger->next;
   }

   return -1;
}

int removeNext(list *theList, element *theElement)
{
   element *finger;
   if(theList == NULL || theElement == NULL || theElement->next == NULL)
   {
      errno = EBAD_LIST;
      return -1;
   }
   finger = theElement->next;
   theElement->next = theElement->next->next;
   free(finger);
   return 0;
}


int findElement(list *theList, void *value, element **retVal)
{
   element *finger;
   if(theList == NULL)
   {
      errno = EBAD_LIST;
      return -1;
   }
   *retVal = NULL;

   finger = theList->head;
   while(finger != NULL)
   {
      if(finger->value == value)
      {
         *retVal = finger;
         return 0;
      }
      finger = finger->next;
   }
   return -1;
}

int headElement(list *theList, element **retVal)
{
   if(theList == NULL)
   {
      errno = EBAD_LIST;
      return -1;
   }
   *retVal = theList->head;
   if(*retVal == NULL)
   {
      return -1;
   }
   return 0;
}

int nextElement(list *theList, element *handle, element **retVal)
{
   if(theList == NULL)
   {
      errno = EBAD_LIST;
      return -1;
   }
   if(handle == NULL)
   {
      errno = EBAD_HANDLE;
      return -1;
   }
   *retVal = handle->next;
   if(*retVal == NULL)
   {
      return -1;
   }
   return 0;
}

int prevElement(list *theList, element *handle, element **retVal)
{
   if(theList == NULL)
   {
      errno = EBAD_LIST;
      return -1;
   }
   /* implement later if neccessary */
   return 0;
}

int listSize(list *theList)
{
   if(theList == NULL)
   {
      errno = EBAD_LIST;
      return -1;
   }
   return theList->size;
}
