/*  Written by Julian Grizzard
 *  September 2, 2002
 *  Contact: gtg993j@mail.gatech.edu
 *
 *  Please contact me if you decide to publish any of this code.  Thanks!
 */

#ifndef _LINKED_LIST_H
#define _LINKED_LIST_H

/* structs */

typedef struct list
{
   struct element *head;
   int size;
} list;

typedef struct element
{
   void *value;
   struct element *next;
} element;

/* errors */

#define EBAD_LIST 300
#define EBAD_HANDLE 301

/* Returns a new initialized list
 * or NULL if error
 */
list *newList();

/* Returns a copy of list or
 * NULL if list is NULL
 */
list *copyList(list *theList);

/* frees a list */
int freeList(list *theList);

/* Wraps value in an element and adds to tail
 * of list or returns -1
 * If element already exists, then it will not
 * ge added again and 0 will be returned.
 */
int addElement(list *theList, void *value);

/* Removes element containing value
 * or returns -1 if not found
 */
int removeElement(list *theList, void *value);

/* Removes the element pointed to after
 * element 
 * returns -1 if next element is NULL
 */
int removeNext(list *theList, element *theElement);

/* returns 0 if element found and return value in retVal
 * returns -1 if not found and null in retVal
 */
int findElement(list *theList, void *value, element **retVal);

/* returns 0 and the head value in retVal (may be NULL)
 * if no head value returns -1
 */
int headElement(list *theList, element **retVal);

/* given an element, returns the next element in the list
 * or NULL and -1 if at end of list
 */
int nextElement(list *theList, element *handle, element **retVal);

/* given an element, returns the previous element in the list
 * or NULL and -1 if at head of list
 */
int prevElement(list *theList, element *handle, element **retVal);

/* returns size of list
 */
int listSize(list *theList);

#endif
