/*  Written by Julian Grizzard
 *  September 2, 2002
 *  Contact: gtg993j@mail.gatech.edu
 *
 *  Please contact me if you decide to publish any of this code.  Thanks!
 */


#ifndef _ROBOT_H
#define _ROBOT_H

#include "linkedList.h"
#include "world.h"

/* structs */

/* define the robot struct */
typedef struct s_robot
{
   int id;        /* Robot id */
   int x;         /* location in x direction */
   int y;         /* location in y direction */
   int money;     /* ammount of money left */
   int cpty;      /* maximum weight capacity for packages */
   list *pkgs;    /* linked list of all carried packages or NULL */
   int state;     /* stores the state of the robot: dead, alive, etc */
} s_robot;

/* commands */
#define C_INIT "Player"
#define C_MOVE "Move "
#define C_PICK "Pick"
#define C_DROP "Drop"
#define C_MAX_MOVE_SIZE 1

/* actions */
#define A_NORTH 'N'
#define A_NORTH_S "N"
#define A_SOUTH 'S'
#define A_SOUTH_S "S"
#define A_EAST 'E'
#define A_EAST_S "E"
#define A_WEST 'W'
#define A_WEST_S "W"
#define A_PICK 'P'
#define A_DROP 'D'
#define A_MOVE 'M'
#define A_XCORD 'X'
#define A_YCORD 'Y'
#define A_ID '#'

/* board */
#define B_PLAIN '.'
#define B_WATER '~'
#define B_WALL '#'
#define B_HOME '@'

/* robot states */
#define R_DEAD 0
#define R_ALIVE 1

/* errors */
#define ENOMEM 100
#define EBAD_BOT 101

/* Send command to move in direction d
 * where d is N, S, E, or W
 */
int move(char *d, int priority, list *robotL); 

/* Send command to pick up list where list is a space seperated
 * list of packages
 */
int pick(char *pickList, int priority, list *robotL);

/* Send command to drop list where list is a space seperated
 * list of packages
 */
int drop(char *dropList, int priority, list *robotL);

/* Creates and return new robot or NULL */
s_robot *newBot();

/* Find robot with id robotID and return pointer to it */
s_robot *findBot(list *robotL, int robotID);

/* Add a package to robot X */
int addPkg(s_robot *robot, list *pkgs, int pkgID);

/* Drop a package from robot X
 * Remove from package list if it hit a target
 */
int dropPkg(s_robot *robot, list *pkgs, int pkgID);


/* Print everything */
void printState(s_world *world, list *robotL, list *homeL, list *pkgL);

/* Print the world */
void printWorld(s_world *world, list *robotL);

/* Print the Robots */
void printRobotL(list *robotL);

/* Print the homes */
void printHomeL(list *homeL);

/* Print the Packages */
void printPkgL(list *pkgL);

/* get my robot x loc */
int botX(list *robotL);

/* get my robot y loc */
int botY(list *robotL);

/* if robot at x,y then return 1, else 0 or -1 for error */
int robotAt(list *robotL, int x, int y);

#endif
